#ifndef _PARSE_SI_H
#define _PARSE_SI_H

HX_BOOL set_eit_language_code(char *letter);
void parse_short_event_descriptor(const unsigned char *buf, struct eit_info_st *eit_info);
void parse_extended_event_descriptor(const unsigned char *buf, struct eit_info_st *eit_info);

/////////////////// TABLE //////////////////////////////////////////
int parse_eit_si(struct parse_info_st *parse_info);

/////////////////// DESCRIPTOR //////////////////////////////////////////
void parse_parental_rating_descriptor (const unsigned char *buf, struct eit_info_st *eit_info);
void parse_content_descriptor(const unsigned char *buf, struct eit_info_st *eit_info);
void parse_component_descriptor(const unsigned char *buf, struct eit_info_st *eit_info);
void parse_linkage_descriptor(const unsigned char *buf, struct eit_info_st *eit_info);

/////////////////////////////////////////////////////////////////////

HX_STATUS
parse_si_section(struct parse_info_st *parse_info, struct service_st *service);

#endif // _PARSE_SI_H
